!MNH_LIC Copyright 1994-2014 CNRS, Meteo-France and Universite Paul Sabatier
!MNH_LIC This is part of the Meso-NH software governed by the CeCILL-C licence
!MNH_LIC version 1. See LICENSE, CeCILL-C_V1-en.txt and CeCILL-C_V1-fr.txt  
!MNH_LIC for details. version 1.
!-----------------------------------------------------------------
!--------------- special set of characters for RCS information
!-----------------------------------------------------------------
! $Source$ $Revision$
! MASDEV4_7 turb 2006/05/18 13:07:25
!-----------------------------------------------------------------
!    #######################
     MODULE MODI_TURB_HOR_UW
!    #######################
!
INTERFACE  
!
      SUBROUTINE TURB_HOR_UW(KSPLT,                                  &
                      OCLOSE_OUT,OTURB_FLX,KRR,                      &
                      HFMFILE,HLUOUT,                                &
                      PK,PINV_PDXX,PINV_PDZZ,PMZM_PRHODJ,            &
                      PDXX,PDZZ,PDZX,                                &
                      PRHODJ,PTHVREF,                                &
                      PUM,PWM,PTHLM,PRM,PSVM,                        &
                      PTKEM,PLM,                                     &
                      PDP,                                           &
                      PRUS,PRWS                                      )
!

!
INTEGER,                  INTENT(IN)    ::  KSPLT        ! split process index
LOGICAL,                  INTENT(IN)    ::  OCLOSE_OUT   ! switch for syncronous
                                                         ! file opening       
LOGICAL,                  INTENT(IN)    ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
INTEGER,                  INTENT(IN)    ::  KRR          ! number of moist var.
CHARACTER(LEN=*),         INTENT(IN)    ::  HFMFILE      ! Name of the output
                                                         ! FM-file 
CHARACTER(LEN=*),         INTENT(IN)    ::  HLUOUT       ! Output-listing name
                                                         ! for model n
!

REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PK          ! Turbulent diffusion doef.
                                                        ! PK = PLM * SQRT(PTKEM)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDXX   ! 1./PDXX
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDZZ   ! 1./PDZZ
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PMZM_PRHODJ ! MZM(PRHODJ)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PDXX, PDZZ, PDZX
                                                         ! Metric coefficients
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PRHODJ       ! density * grid volume
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PTHVREF      ! ref. state VPT       
!
! Variables at t-1
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PUM,PWM,PTHLM
REAL, DIMENSION(:,:,:,:), INTENT(IN)    ::  PRM
REAL, DIMENSION(:,:,:,:), INTENT(IN)    ::  PSVM
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PTKEM        ! TKE at time t- dt
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PLM          ! Turb. mixing length
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PRUS, PRWS
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PDP          ! TKE production terms
!
!$acc declare present(PK,PINV_PDXX,PINV_PDZZ,PMZM_PRHODJ,PDXX,PDZZ,PDZX, &
!$acc &               PRHODJ,PTHVREF,                                    &
!$acc &               PUM,PWM,PTHLM,PRM,PSVM,PTKEM,PLM,                  &
!$acc &               PRUS,PRWS,PDP)
!
END SUBROUTINE TURB_HOR_UW
!
END INTERFACE
!
END MODULE MODI_TURB_HOR_UW
!     ################################################################
      SUBROUTINE TURB_HOR_UW(KSPLT,                                  &
                      OCLOSE_OUT,OTURB_FLX,KRR,                      &
                      HFMFILE,HLUOUT,                                &
                      PK,PINV_PDXX,PINV_PDZZ,PMZM_PRHODJ,            &
                      PDXX,PDZZ,PDZX,                                &
                      PRHODJ,PTHVREF,                                &
                      PUM,PWM,PTHLM,PRM,PSVM,                        &
                      PTKEM,PLM,                                     &
                      PDP,                                           &
                      PRUS,PRWS                                      )
!     ################################################################
!
!
!!****  *TURB_HOR* -routine to compute the source terms in the meso-NH
!!               model equations due to the non-vertical turbulent fluxes.
!!
!!    PURPOSE
!!    -------
!!
!!     see TURB_HOR
!!
!!**  METHOD
!!    ------
!!
!!    EXTERNAL
!!    --------
!!
!!    IMPLICIT ARGUMENTS
!!    ------------------
!!
!!    REFERENCE
!!    ---------
!!
!!    AUTHOR
!!    ------
!!
!!      Joan Cuxart             * INM and Meteo-France *
!!
!!    MODIFICATIONS
!!    -------------
!!                     Aug    , 1997 (V. Saravane) spliting of TURB_HOR
!!                     Nov  27, 1997 (V. Masson) clearing of the routine
!!                     Oct  18, 2000 (V. Masson) LES computations + LFLAT switch
!!                     Feb  14, 2001 (V. Masson and J. Stein) DZF bug on PRWS
!!                                   + remove the use of W=0 at the ground
!!                                   + extrapolation under the ground
!!                     Nov  06, 2002 (V. Masson) LES budgets
!!                     October 2009 (G. Tanguy) add ILENCH=LEN(YCOMMENT) after
!!                                              change of YCOMMENT
!!                     04/2016 (M.Moge) Use openACC directives to port the TURB part of Meso-NH on GPU
!! --------------------------------------------------------------------------
!
!*      0. DECLARATIONS
!          ------------
!
USE MODD_CST
USE MODD_CONF
USE MODD_CTURB
USE MODD_PARAMETERS
USE MODD_LES
USE MODD_NSV
!
USE MODE_FMWRIT
USE MODI_GRADIENT_M
USE MODI_GRADIENT_U
USE MODI_GRADIENT_V
USE MODI_GRADIENT_W
#ifndef _OPENACC
USE MODI_SHUMAN
#else
USE MODI_SHUMAN_DEVICE
#endif
USE MODI_COEFJ
USE MODI_LES_MEAN_SUBGRID
!
USE MODI_SECOND_MNH
!
IMPLICIT NONE
!
!
!*       0.1  declaration of arguments
!
!
!
INTEGER,                  INTENT(IN)    ::  KSPLT        ! split process index
LOGICAL,                  INTENT(IN)    ::  OCLOSE_OUT   ! switch for syncronous
                                                         ! file opening       
LOGICAL,                  INTENT(IN)    ::  OTURB_FLX    ! switch to write the
                                 ! turbulent fluxes in the syncronous FM-file
INTEGER,                  INTENT(IN)    ::  KRR          ! number of moist var.
CHARACTER(LEN=*),         INTENT(IN)    ::  HFMFILE      ! Name of the output
                                                         ! FM-file 
CHARACTER(LEN=*),         INTENT(IN)    ::  HLUOUT       ! Output-listing name
                                                         ! for model n
!

REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PK          ! Turbulent diffusion doef.
                                                        ! PK = PLM * SQRT(PTKEM)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDXX   ! 1./PDXX
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PINV_PDZZ   ! 1./PDZZ
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PMZM_PRHODJ ! MZM(PRHODJ)
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PDXX, PDZZ, PDZX
                                                         ! Metric coefficients
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PRHODJ       ! density * grid volume
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PTHVREF      ! ref. state VPT       
!
! Variables at t-1
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PUM,PWM,PTHLM
REAL, DIMENSION(:,:,:,:), INTENT(IN)    ::  PRM
REAL, DIMENSION(:,:,:,:), INTENT(IN)    ::  PSVM
!
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PTKEM        ! TKE at time t- dt
REAL, DIMENSION(:,:,:),   INTENT(IN)    ::  PLM          ! Turb. mixing length
!
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PRUS, PRWS
REAL, DIMENSION(:,:,:),   INTENT(INOUT) ::  PDP          ! TKE production terms
!$acc declare present(PK,PINV_PDXX,PINV_PDZZ,PMZM_PRHODJ,PDXX,PDZZ,PDZX, &
!$acc &               PRHODJ,PTHVREF,                                    &
!$acc &               PUM,PWM,PTHLM,PRM,PSVM,PTKEM,PLM,                  &
!$acc &               PRUS,PRWS,PDP)
!
!*       0.2  declaration of local variables
!
REAL, DIMENSION(SIZE(PWM,1),SIZE(PWM,2),SIZE(PWM,3))       &
                                     :: ZFLX,ZWORK
    ! work arrays
!   
INTEGER             :: IRESP        ! Return code of FM routines 
INTEGER             :: IGRID        ! C-grid indicator in LFIFM file 
INTEGER             :: ILENCH       ! Length of comment string in LFIFM file
INTEGER             :: IKB,IKE,IKU
                                    ! Index values for the Beginning and End
                                    ! mass points of the domain  
INTEGER             :: JSV          ! scalar loop counter
CHARACTER (LEN=100) :: YCOMMENT     ! comment string in LFIFM file
CHARACTER (LEN=16)  :: YRECFM       ! Name of the desired field in LFIFM file
!
REAL, DIMENSION(SIZE(PWM,1),SIZE(PWM,2),SIZE(PWM,3))  :: GX_W_UW_PWM
!$acc declare create(ZFLX,ZWORK,GX_W_UW_PWM)
!
REAL :: ZTIME1, ZTIME2
!
#ifdef _OPENACC
REAL, DIMENSION(SIZE(PWM,1),SIZE(PWM,2),SIZE(PWM,3)) :: ZTMP1_DEVICE
REAL, DIMENSION(SIZE(PWM,1),SIZE(PWM,2),SIZE(PWM,3)) :: ZTMP2_DEVICE
REAL, DIMENSION(SIZE(PWM,1),SIZE(PWM,2),SIZE(PWM,3)) :: ZTMP3_DEVICE
REAL, DIMENSION(SIZE(PWM,1),SIZE(PWM,2),SIZE(PWM,3)) :: ZTMP4_DEVICE
!$acc declare create(ZTMP1_DEVICE,ZTMP2_DEVICE,ZTMP3_DEVICE,ZTMP4_DEVICE)
#endif
! ---------------------------------------------------------------------------
!
!*       1.   PRELIMINARY COMPUTATIONS
!             ------------------------
!
IKB = 1+JPVEXT               
IKE = SIZE(PWM,3)-JPVEXT    
IKU = SIZE(PWM,3)
!
!
#ifndef _OPENACC
GX_W_UW_PWM = GX_W_UW(1,IKU,1,PWM,PDXX,PDZZ,PDZX)
#else
CALL GX_W_UW_DEVICE(1,IKU,1,PWM,PDXX,PDZZ,PDZX,GX_W_UW_PWM)
#endif
!
!
!*      13.   < U'W'>
!             -------
! 
! residual part of < U'W'> depending on dw/dx
!
#ifndef _OPENACC
ZFLX(:,:,:) =                                                      &
  - XCMFS * MXM(MZM(1,IKU,1,PK)) * GX_W_UW_PWM
#else
CALL MZM_DEVICE(PK,ZTMP1_DEVICE)
CALL MXM_DEVICE(ZTMP1_DEVICE,ZTMP2_DEVICE)
!$acc kernels
ZFLX(:,:,:) = - XCMFS * ZTMP2_DEVICE * GX_W_UW_PWM
#endif
!!         &  to be tested
!!  - (2./3.) * XCMFB * MZM( ZVPTU * MXM( PLM / SQRT(PTKEM) * XG / PTHVREF ) )
!
ZFLX(:,:,IKE+1) = 0.  ! rigid wall condition => no turbulent flux
!
! Nullify the flux at the ground level because it has been fully taken into
! account in turb_ver and extrapolate the flux under the ground 
ZFLX(:,:,IKB) = 0.
ZFLX(:,:,IKB-1)=2.*ZFLX(:,:,IKB)- ZFLX(:,:,IKB+1)
!$acc end kernels
!
! stores  <U W>
IF ( OCLOSE_OUT .AND. OTURB_FLX ) THEN
!$acc update self(ZFLX)
  YRECFM  ='UW_HFLX'
  YCOMMENT='X_Y_Z_UW_HFLX ( (M/S) **2 )  '
  IGRID   = 6
  ILENCH=LEN(YCOMMENT)
  CALL FMWRIT(HFMFILE,YRECFM,HLUOUT,'XY',ZFLX,IGRID,ILENCH,YCOMMENT,IRESP)
END IF
!
!
! compute the source for rho*U due to this residual flux ( the other part is
! taken into account in TURB_VER)
#ifndef _OPENACC
PRUS(:,:,:) = PRUS(:,:,:) - DZF(1,IKU,1, ZFLX* MXM( PMZM_PRHODJ ) / MXM( PDZZ ) )
#else
CALL MXM_DEVICE( PMZM_PRHODJ, ZTMP1_DEVICE )
CALL MXM_DEVICE( PDZZ, ZTMP2_DEVICE )
!$acc kernels
ZTMP3_DEVICE = ZFLX* ZTMP1_DEVICE / ZTMP2_DEVICE
!$acc end kernels
CALL DZF_DEVICE(1,IKU,1, ZTMP3_DEVICE, ZTMP1_DEVICE )
!$acc kernels
PRUS(:,:,:) = PRUS(:,:,:) - ZTMP1_DEVICE(:,:,:)
!$acc end kernels
#endif
!
!computation of the source for rho*W due to this flux
#ifndef _OPENACC
IF (.NOT. LFLAT) THEN
  PRWS(:,:,:) = PRWS(:,:,:)                              &
        -DXF( MZM(1,IKU,1, MXM(PRHODJ) * PINV_PDXX) * ZFLX)           &
        +DZM(1,IKU,1, PRHODJ * MXF( MZF(1,IKU,1, ZFLX*PDZX ) * PINV_PDXX ) / MZF(1,IKU,1,PDZZ) )
ELSE
  PRWS(:,:,:) = PRWS(:,:,:) -DXF( MZM(1,IKU,1, MXM(PRHODJ) * PINV_PDXX) * ZFLX)
END IF
#else
  CALL MXM_DEVICE(PRHODJ,ZTMP1_DEVICE)
  !$acc kernels
  ZTMP2_DEVICE = ZTMP1_DEVICE * PINV_PDXX
  !$acc end kernels
  CALL MZM_DEVICE(ZTMP2_DEVICE, ZTMP1_DEVICE)
  !$acc kernels
  ZTMP2_DEVICE = ZTMP1_DEVICE * ZFLX
  !$acc end kernels
  CALL DXF_DEVICE( ZTMP2_DEVICE,ZTMP1_DEVICE)
IF (.NOT. LFLAT) THEN
  !$acc kernels
  ZTMP2_DEVICE = ZFLX*PDZX
  !$acc end kernels
  CALL MZF_DEVICE(1,IKU,1, ZTMP2_DEVICE, ZTMP3_DEVICE )
  !$acc kernels
  ZTMP2_DEVICE = ZTMP3_DEVICE*PINV_PDXX
  !$acc end kernels
  CALL MXF_DEVICE( ZTMP2_DEVICE, ZTMP3_DEVICE )
  CALL MZF_DEVICE(1,IKU,1,PDZZ, ZTMP2_DEVICE)
  !$acc kernels
  ZTMP4_DEVICE = PRHODJ * ZTMP3_DEVICE / ZTMP2_DEVICE
  !$acc end kernels
  CALL DZM_DEVICE(1,IKU,1, ZTMP4_DEVICE, ZTMP2_DEVICE )
  !$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:)                              &
        - ZTMP1_DEVICE           &
        + ZTMP2_DEVICE
  !$acc end kernels
ELSE 
  !$acc kernels
  PRWS(:,:,:) = PRWS(:,:,:) - ZTMP1_DEVICE
  !$acc end kernels
END IF
#endif
! 
IF (KSPLT==1) THEN
  !
  !Contribution to the dynamic production of TKE:
  !
#ifndef _OPENACC
  ZWORK(:,:,:) =-MZF(1,IKU,1, MXF(                               &
     ZFLX *( GZ_U_UW(1,IKU,1,PUM,PDZZ) + GX_W_UW_PWM ) ) )
#else
  CALL GZ_U_UW_DEVICE(1,IKU,1,PUM,PDZZ,ZTMP1_DEVICE)
  !$acc kernels
  ZTMP2_DEVICE = ZFLX *( ZTMP1_DEVICE + GX_W_UW_PWM )
  !$acc end kernels
  CALL MXF_DEVICE( ZTMP2_DEVICE,ZTMP1_DEVICE )
  CALL MZF_DEVICE(1,IKU,1, ZTMP1_DEVICE, ZTMP2_DEVICE )
  !$acc kernels
  ZWORK(:,:,:) = -ZTMP2_DEVICE(:,:,:)
  !$acc end kernels
#endif
  !
  !
  ! evaluate the dynamic production at w(IKB+1) in PDP(IKB)
#ifndef _OPENACC
  ZWORK(:,:,IKB:IKB) = - MXF (                                             &
     ZFLX(:,:,IKB+1:IKB+1) *                                               &
   (   (PUM(:,:,IKB+1:IKB+1)-PUM(:,:,IKB:IKB)) / MXM(PDZZ(:,:,IKB+1:IKB+1))&
     + ( DXM( PWM(:,:,IKB+1:IKB+1) )                                       &
        -MXM(  (PWM(:,:,IKB+2:IKB+2)-PWM(:,:,IKB+1:IKB+1))                 &
                /(PDZZ(:,:,IKB+2:IKB+2)+PDZZ(:,:,IKB+1:IKB+1))             &
              +(PWM(:,:,IKB+1:IKB+1)-PWM(:,:,IKB  :IKB  ))                 &
                /(PDZZ(:,:,IKB+1:IKB+1)+PDZZ(:,:,IKB  :IKB  ))             &
            )                                                              &
          * PDZX(:,:,IKB+1:IKB+1)                                          &
       ) / (0.5*(PDXX(:,:,IKB+1:IKB+1)+PDXX(:,:,IKB:IKB)))                 &
   )                        )
#else
  CALL MXM_DEVICE( PDZZ(:,:,IKB+1:IKB+1), ZTMP1_DEVICE(:,:,1:1) )
  CALL DXM_DEVICE( PWM(:,:,IKB+1:IKB+1), ZTMP2_DEVICE(:,:,1:1) )
  !$acc kernels
  ZTMP3_DEVICE(:,:,1) = (PWM(:,:,IKB+2)-PWM(:,:,IKB+1)) &
                /(PDZZ(:,:,IKB+2)+PDZZ(:,:,IKB+1))      &
              +(PWM(:,:,IKB+1)-PWM(:,:,IKB))            &
                /(PDZZ(:,:,IKB+1)+PDZZ(:,:,IKB))
  !$acc end kernels
  CALL MXM_DEVICE( ZTMP3_DEVICE(:,:,1:1), ZTMP4_DEVICE(:,:,1:1) )
  !$acc kernels
  ZTMP3_DEVICE(:,:,1) = ZFLX(:,:,IKB+1) *                  &
   (   (PUM(:,:,IKB+1)-PUM(:,:,IKB)) / ZTMP1_DEVICE(:,:,1) &
     + ( ZTMP2_DEVICE(:,:,1)                               &
       - ZTMP4_DEVICE(:,:,1)                               &
       * PDZX(:,:,IKB+1)                                   &
       ) / (0.5*(PDXX(:,:,IKB+1)+PDXX(:,:,IKB)))           &
   )
  !$acc end kernels
  CALL MXF_DEVICE( ZTMP3_DEVICE(:,:,1:1), ZTMP1_DEVICE(:,:,1:1) )
  !$acc kernels
  ZWORK(:,:,IKB) = - ZTMP1_DEVICE(:,:,1) 
  !$acc end kernels
#endif
  !
  ! dynamic production computation
  !$acc kernels
  PDP(:,:,:) = PDP(:,:,:) +  ZWORK(:,:,:)  
  !$acc end kernels
  !
END IF
!
! Storage in the LES configuration (addition to TURB_VER computation)
!
IF (LLES_CALL .AND. KSPLT==1) THEN
  CALL SECOND_MNH(ZTIME1)
#ifndef _OPENACC
  CALL LES_MEAN_SUBGRID( MZF(1,IKU,1,MXF(ZFLX)), X_LES_SUBGRID_WU , .TRUE. ) 
  CALL LES_MEAN_SUBGRID( MZF(1,IKU,1,MXF(GZ_U_UW(1,IKU,1,PUM,PDZZ)*ZFLX)), X_LES_RES_ddxa_U_SBG_UaU , .TRUE.)
  CALL LES_MEAN_SUBGRID( MZF(1,IKU,1,MXF(GX_W_UW_PWM*ZFLX)), X_LES_RES_ddxa_W_SBG_UaW , .TRUE.)
  CALL LES_MEAN_SUBGRID( MXF(GX_M_U(1,IKU,1,PTHLM,PDXX,PDZZ,PDZX)*MZF(1,IKU,1,ZFLX)),&
                         X_LES_RES_ddxa_Thl_SBG_UaW , .TRUE.)
  IF (KRR>=1) THEN
    CALL LES_MEAN_SUBGRID( MXF(GX_M_U(1,IKU,1,PRM(:,:,:,1),PDXX,PDZZ,PDZX)*MZF(1,IKU,1,ZFLX)), &
                           X_LES_RES_ddxa_Rt_SBG_UaW , .TRUE.)
  END IF
  DO JSV=1,NSV
    CALL LES_MEAN_SUBGRID( MXF(GX_M_U(1,IKU,1,PSVM(:,:,:,JSV),PDXX,PDZZ,PDZX)*MZF(1,IKU,1,ZFLX)), &
                           X_LES_RES_ddxa_Sv_SBG_UaW(:,:,:,JSV) , .TRUE.)
  END DO
#else
!$acc data copy(X_LES_SUBGRID_WU,X_LES_RES_ddxa_U_SBG_UaU,X_LES_RES_ddxa_W_SBG_UaW, &
!$acc &         X_LES_RES_ddxa_Thl_SBG_UaW,X_LES_RES_ddxa_Sv_SBG_UaW)
  !
  CALL MXF_DEVICE(ZFLX,ZTMP1_DEVICE)
  CALL MZF_DEVICE(1,IKU,1,ZTMP1_DEVICE,ZTMP2_DEVICE)
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_SUBGRID_WU , .TRUE. ) 
  !
  CALL GZ_U_UW_DEVICE(1,IKU,1,PUM,PDZZ,ZTMP1_DEVICE)
  !$acc kernels
  ZTMP2_DEVICE = ZTMP1_DEVICE*ZFLX
  !$acc end kernels
  CALL MXF_DEVICE(ZTMP2_DEVICE,ZTMP1_DEVICE)
  CALL MZF_DEVICE(1,IKU,1,ZTMP1_DEVICE,ZTMP2_DEVICE)
  CALL LES_MEAN_SUBGRID( ZTMP2_DEVICE, X_LES_RES_ddxa_U_SBG_UaU , .TRUE.)
  !
  !$acc kernels
  ZTMP1_DEVICE = GX_W_UW_PWM*ZFLX
  !$acc end kernels
  CALL MXF_DEVICE(ZTMP1_DEVICE,ZTMP2_DEVICE)
  CALL MZF_DEVICE(1,IKU,1,ZTMP2_DEVICE,ZTMP1_DEVICE)
  CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_RES_ddxa_W_SBG_UaW , .TRUE.)
  !
  CALL MZF_DEVICE(1,IKU,1,ZFLX,ZTMP1_DEVICE)
  CALL GX_M_U_DEVICE(1,IKU,1,PTHLM,PDXX,PDZZ,PDZX,ZTMP2_DEVICE)
  !$acc kernels
  ZTMP3_DEVICE = ZTMP2_DEVICE*ZTMP1_DEVICE
  !$acc end kernels
  CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP1_DEVICE)
  CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE,X_LES_RES_ddxa_Thl_SBG_UaW , .TRUE.)
  !
  IF (KRR>=1) THEN
    !
!$acc data copy(X_LES_RES_ddxa_Rt_SBG_UaW)
    !
    CALL MZF_DEVICE(1,IKU,1,ZFLX,ZTMP1_DEVICE)
    CALL GX_M_U_DEVICE(1,IKU,1,PRM(:,:,:,1),PDXX,PDZZ,PDZX,ZTMP2_DEVICE)
    !$acc kernels
    ZTMP3_DEVICE = ZTMP1_DEVICE*ZTMP2_DEVICE
    !$acc end kernels
    CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP1_DEVICE)
    CALL LES_MEAN_SUBGRID(ZTMP1_DEVICE, X_LES_RES_ddxa_Rt_SBG_UaW , .TRUE.)
    !
!$acc end data
    !
  END IF
  !
  CALL MZF_DEVICE(1,IKU,1,ZFLX,ZTMP1_DEVICE)
  DO JSV=1,NSV
    CALL GX_M_U_DEVICE(1,IKU,1,PSVM(:,:,:,JSV),PDXX,PDZZ,PDZX,ZTMP2_DEVICE)
    !$acc kernels
    ZTMP3_DEVICE = ZTMP1_DEVICE*ZTMP2_DEVICE
    !$acc end kernels
    CALL MXF_DEVICE(ZTMP3_DEVICE,ZTMP1_DEVICE)
    CALL LES_MEAN_SUBGRID( ZTMP1_DEVICE, X_LES_RES_ddxa_Sv_SBG_UaW(:,:,:,JSV) , .TRUE.)
  END DO
  !
!$acc end data
#endif
  CALL SECOND_MNH(ZTIME2)
  XTIME_LES = XTIME_LES + ZTIME2 - ZTIME1
END IF

!
END SUBROUTINE TURB_HOR_UW
